@@ -4,8 +4,8 @@ import (
 	"bytes"
 	"container/heap"
 	"encoding/hex"
+	"fmt"
 	"io"
-	"io/ioutil"
 	"log"
 	"net/http"
 	"net/http/httptest"
@@ -14,24 +14,70 @@ import (
 	"testing"
 	"time"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/amp"
 	. "github.com/smartystreets/goconvey/convey"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/amp"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/messages"
 )
 
 func NullLogger() *log.Logger {
 	logger := log.New(os.Stdout, "", 0)
-	logger.SetOutput(ioutil.Discard)
+	logger.SetOutput(io.Discard)
 	return logger
 }
 
 var promOnce sync.Once
 
+var (
+	sdp = "v=0\r\n" +
+		"o=- 123456789 987654321 IN IP4 0.0.0.0\r\n" +
+		"s=-\r\n" +
+		"t=0 0\r\n" +
+		"a=fingerprint:sha-256 12:34\r\n" +
+		"a=extmap-allow-mixed\r\n" +
+		"a=group:BUNDLE 0\r\n" +
+		"m=application 9 UDP/DTLS/SCTP webrtc-datachannel\r\n" +
+		"c=IN IP4 0.0.0.0\r\n" +
+		"a=setup:actpass\r\n" +
+		"a=mid:0\r\n" +
+		"a=sendrecv\r\n" +
+		"a=sctp-port:5000\r\n" +
+		"a=ice-ufrag:CoVEaiFXRGVzshXG\r\n" +
+		"a=ice-pwd:aOrOZXraTfFKzyeBxIXYYKjSgRVPGhUx\r\n" +
+		"a=candidate:1000 1 udp 2000 8.8.8.8 3000 typ host\r\n" +
+		"a=end-of-candidates\r\n"
+
+	sid = "ymbcCMto7KHNGYlp"
+)
+
+func createClientOffer(sdp, nat, fingerprint string) (*bytes.Reader, error) {
+	clientRequest := &messages.ClientPollRequest{
+		Offer:       sdp,
+		NAT:         nat,
+		Fingerprint: fingerprint,
+	}
+	encOffer, err := clientRequest.EncodeClientPollRequest()
+	if err != nil {
+		return nil, err
+	}
+	offer := bytes.NewReader(encOffer)
+	return offer, nil
+}
+
+func createProxyAnswer(sdp, sid string) (*bytes.Reader, error) {
+	proxyRequest, err := messages.EncodeAnswerRequest(sdp, sid)
+	if err != nil {
+		return nil, err
+	}
+	answer := bytes.NewReader(proxyRequest)
+	return answer, nil
+}
+
 func decodeAMPArmorToString(r io.Reader) (string, error) {
 	dec, err := amp.NewArmorDecoder(r)
 	if err != nil {
 		return "", err
 	}
-	p, err := ioutil.ReadAll(dec)
+	p, err := io.ReadAll(dec)
 	return string(p), err
 }
 
@@ -42,7 +88,8 @@ func TestBroker(t *testing.T) {
 	copy(defaultBridge[:], defaultBridgeValue)
 
 	Convey("Context", t, func() {
-		ctx := NewBrokerContext(NullLogger())
+		buf := new(bytes.Buffer)
+		ctx := NewBrokerContext(log.New(buf, "", 0), "", "")
 		i := &IPC{ctx}
 
 		Convey("Adds Snowflake", func() {
@@ -86,8 +133,7 @@ func TestBroker(t *testing.T) {
 
 		Convey("Responds to HTTP client offers...", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unknown\"}"))
+			data, err := createClientOffer(sdp, NATUnknown, "")
 			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 
@@ -95,22 +141,74 @@ func TestBroker(t *testing.T) {
 				clientOffers(i, w, r)
 				So(w.Code, ShouldEqual, http.StatusOK)
 				So(w.Body.String(), ShouldEqual, `{"error":"no snowflake proxies currently available"}`)
+
+				// Ensure that denial is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 8
+client-restricted-denied-count 8
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 8
+client-http-ips ??=8
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+`)
 			})
 
 			Convey("with a proxy answer if available.", func() {
 				done := make(chan bool)
 				// Prepare a fake proxy to respond with.
-				snowflake := ctx.AddSnowflake("fake", "", NATUnrestricted, 0)
+				snowflake := ctx.AddSnowflake("test", "", NATUnrestricted, 0)
 				go func() {
 					clientOffers(i, w, r)
 					done <- true
 				}()
 				offer := <-snowflake.offerChannel
-				So(offer.sdp, ShouldResemble, []byte("fake"))
-				snowflake.answerChannel <- "fake answer"
+				So(offer.sdp, ShouldResemble, []byte(sdp))
+				snowflake.answerChannel <- "test answer"
 				<-done
-				So(w.Body.String(), ShouldEqual, `{"answer":"fake answer"}`)
+				So(w.Body.String(), ShouldEqual, `{"answer":"test answer"}`)
 				So(w.Code, ShouldEqual, http.StatusOK)
+
+				// Ensure that match is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 0
+client-restricted-denied-count 0
+client-unrestricted-denied-count 0
+client-snowflake-match-count 8
+client-http-count 8
+client-http-ips ??=8
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+`)
+			})
+
+			Convey("with unrestricted proxy to unrestricted client if there are no restricted proxies", func() {
+				snowflake := ctx.AddSnowflake("test", "", NATUnrestricted, 0)
+				offerData, err := createClientOffer(sdp, NATUnrestricted, "")
+				So(err, ShouldBeNil)
+				r, err := http.NewRequest("POST", "snowflake.broker/client", offerData)
+
+				done := make(chan bool)
+				go func() {
+					clientOffers(i, w, r)
+					done <- true
+				}()
+
+				select {
+				case <-snowflake.offerChannel:
+				case <-time.After(250 * time.Millisecond):
+					So(false, ShouldBeTrue)
+					return
+				}
+				snowflake.answerChannel <- "test answer"
+
+				<-done
+				So(w.Body.String(), ShouldEqual, `{"answer":"test answer"}`)
 			})
 
 			Convey("Times out when no proxy responds.", func() {
@@ -125,7 +223,7 @@ func TestBroker(t *testing.T) {
 					done <- true
 				}()
 				offer := <-snowflake.offerChannel
-				So(offer.sdp, ShouldResemble, []byte("fake"))
+				So(offer.sdp, ShouldResemble, []byte(sdp))
 				<-done
 				So(w.Code, ShouldEqual, http.StatusOK)
 				So(w.Body.String(), ShouldEqual, `{"error":"timed out waiting for answer!"}`)
@@ -134,8 +232,9 @@ func TestBroker(t *testing.T) {
 
 		Convey("Responds to HTTP legacy client offers...", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader([]byte("{test}"))
-			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
+			// legacy offer starts with {
+			offer := bytes.NewReader([]byte(fmt.Sprintf(`{%v}`, sdp)))
+			r, err := http.NewRequest("POST", "snowflake.broker/client", offer)
 			So(err, ShouldBeNil)
 			r.Header.Set("Snowflake-NAT-TYPE", "restricted")
 
@@ -143,6 +242,20 @@ func TestBroker(t *testing.T) {
 				clientOffers(i, w, r)
 				So(w.Code, ShouldEqual, http.StatusServiceUnavailable)
 				So(w.Body.String(), ShouldEqual, "")
+
+				// Ensure that denial is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 8
+client-restricted-denied-count 8
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 8
+client-http-ips ??=8
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+`)
 			})
 
 			Convey("with a proxy answer if available.", func() {
@@ -154,11 +267,25 @@ func TestBroker(t *testing.T) {
 					done <- true
 				}()
 				offer := <-snowflake.offerChannel
-				So(offer.sdp, ShouldResemble, []byte("{test}"))
+				So(offer.sdp, ShouldResemble, []byte(fmt.Sprintf(`{%v}`, sdp)))
 				snowflake.answerChannel <- "fake answer"
 				<-done
 				So(w.Body.String(), ShouldEqual, "fake answer")
 				So(w.Code, ShouldEqual, http.StatusOK)
+
+				// Ensure that match is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 0
+client-restricted-denied-count 0
+client-unrestricted-denied-count 0
+client-snowflake-match-count 8
+client-http-count 8
+client-http-ips ??=8
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+`)
 			})
 
 			Convey("Times out when no proxy responds.", func() {
@@ -173,7 +300,7 @@ func TestBroker(t *testing.T) {
 					done <- true
 				}()
 				offer := <-snowflake.offerChannel
-				So(offer.sdp, ShouldResemble, []byte("{test}"))
+				So(offer.sdp, ShouldResemble, []byte(fmt.Sprintf(`{%v}`, sdp)))
 				<-done
 				So(w.Code, ShouldEqual, http.StatusGatewayTimeout)
 			})
@@ -201,6 +328,20 @@ func TestBroker(t *testing.T) {
 				body, err := decodeAMPArmorToString(w.Body)
 				So(err, ShouldBeNil)
 				So(body, ShouldEqual, `{"error":"no snowflake proxies currently available"}`)
+
+				// Ensure that denial is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 8
+client-restricted-denied-count 8
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 0
+client-http-ips 
+client-ampcache-count 8
+client-ampcache-ips ??=8
+client-sqs-count 0
+client-sqs-ips 
+`)
 			})
 
 			Convey("with a proxy answer if available.", func() {
@@ -219,6 +360,20 @@ func TestBroker(t *testing.T) {
 				So(err, ShouldBeNil)
 				So(body, ShouldEqual, `{"answer":"fake answer"}`)
 				So(w.Code, ShouldEqual, http.StatusOK)
+
+				// Ensure that match is correctly recorded in metrics
+				ctx.metrics.printMetrics()
+				So(buf.String(), ShouldContainSubstring, `client-denied-count 0
+client-restricted-denied-count 0
+client-unrestricted-denied-count 0
+client-snowflake-match-count 8
+client-http-count 0
+client-http-ips 
+client-ampcache-count 8
+client-ampcache-ips ??=8
+client-sqs-count 0
+client-sqs-ips 
+`)
 			})
 
 			Convey("Times out when no proxy responds.", func() {
@@ -281,9 +436,11 @@ func TestBroker(t *testing.T) {
 
 		Convey("Responds to proxy answers...", func() {
 			done := make(chan bool)
-			s := ctx.AddSnowflake("test", "", NATUnrestricted, 0)
+			s := ctx.AddSnowflake(sid, "", NATUnrestricted, 0)
 			w := httptest.NewRecorder()
-			data := bytes.NewReader([]byte(`{"Version":"1.0","Sid":"test","Answer":"test"}`))
+
+			data, err := createProxyAnswer(sdp, sid)
+			So(err, ShouldBeNil)
 
 			Convey("by passing to the client if valid.", func() {
 				r, err := http.NewRequest("POST", "snowflake.broker/answer", data)
@@ -295,19 +452,18 @@ func TestBroker(t *testing.T) {
 				answer := <-s.answerChannel
 				<-done
 				So(w.Code, ShouldEqual, http.StatusOK)
-				So(answer, ShouldResemble, "test")
+				So(answer, ShouldResemble, sdp)
 			})
 
-			Convey("with client gone status if the proxy is not recognized", func() {
-				data = bytes.NewReader([]byte(`{"Version":"1.0","Sid":"invalid","Answer":"test"}`))
+			Convey("with client gone status if the proxy ID is not recognized", func() {
+				data, err := createProxyAnswer(sdp, "invalid")
 				r, err := http.NewRequest("POST", "snowflake.broker/answer", data)
 				So(err, ShouldBeNil)
 				proxyAnswers(i, w, r)
 				So(w.Code, ShouldEqual, http.StatusOK)
-				b, err := ioutil.ReadAll(w.Body)
+				b, err := io.ReadAll(w.Body)
 				So(err, ShouldBeNil)
 				So(b, ShouldResemble, []byte(`{"Status":"client gone"}`))
-
 			})
 
 			Convey("with error if the proxy gives invalid answer", func() {
@@ -331,7 +487,7 @@ func TestBroker(t *testing.T) {
 	})
 
 	Convey("End-To-End", t, func() {
-		ctx := NewBrokerContext(NullLogger())
+		ctx := NewBrokerContext(NullLogger(), "", "")
 		i := &IPC{ctx}
 
 		Convey("Check for client/proxy data race", func() {
@@ -353,7 +509,8 @@ func TestBroker(t *testing.T) {
 
 			// Client offer
 			wc := httptest.NewRecorder()
-			datac := bytes.NewReader([]byte("test"))
+			datac, err := createClientOffer(sdp, NATUnknown, "")
+			So(err, ShouldBeNil)
 			rc, err := http.NewRequest("POST", "snowflake.broker/client", datac)
 			So(err, ShouldBeNil)
 
@@ -367,7 +524,8 @@ func TestBroker(t *testing.T) {
 
 			// Proxy answers
 			wp = httptest.NewRecorder()
-			datap = bytes.NewReader([]byte(`{"Version":"1.0","Sid":"ymbcCMto7KHNGYlp","Answer":"test"}`))
+			datap, err = createProxyAnswer(sdp, sid)
+			So(err, ShouldBeNil)
 			rp, err = http.NewRequest("POST", "snowflake.broker/answer", datap)
 			So(err, ShouldBeNil)
 			go func(i *IPC) {
@@ -405,9 +563,9 @@ func TestBroker(t *testing.T) {
 			So(ctx.idToSnowflake["ymbcCMto7KHNGYlp"], ShouldNotBeNil)
 
 			// Client request blocks until proxy answer arrives.
-			dataC := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unknown\"}"))
 			wC := httptest.NewRecorder()
+			dataC, err := createClientOffer(sdp, NATUnknown, "")
+			So(err, ShouldBeNil)
 			rC, err := http.NewRequest("POST", "snowflake.broker/client", dataC)
 			So(err, ShouldBeNil)
 			go func() {
@@ -417,11 +575,13 @@ func TestBroker(t *testing.T) {
 
 			<-polled
 			So(wP.Code, ShouldEqual, http.StatusOK)
-			So(wP.Body.String(), ShouldResemble, `{"Status":"client match","Offer":"fake","NAT":"unknown","RelayURL":"wss://snowflake.torproject.net/"}`)
-			So(ctx.idToSnowflake["ymbcCMto7KHNGYlp"], ShouldNotBeNil)
+			So(wP.Body.String(), ShouldResemble, fmt.Sprintf(`{"Status":"client match","Offer":%#q,"NAT":"unknown","RelayURL":"wss://snowflake.torproject.net/"}`, sdp))
+			So(ctx.idToSnowflake[sid], ShouldNotBeNil)
+
 			// Follow up with the answer request afterwards
 			wA := httptest.NewRecorder()
-			dataA := bytes.NewReader([]byte(`{"Version":"1.0","Sid":"ymbcCMto7KHNGYlp","Answer":"test"}`))
+			dataA, err := createProxyAnswer(sdp, sid)
+			So(err, ShouldBeNil)
 			rA, err := http.NewRequest("POST", "snowflake.broker/answer", dataA)
 			So(err, ShouldBeNil)
 			proxyAnswers(i, wA, rA)
@@ -429,7 +589,7 @@ func TestBroker(t *testing.T) {
 
 			<-done
 			So(wC.Code, ShouldEqual, http.StatusOK)
-			So(wC.Body.String(), ShouldEqual, `{"answer":"test"}`)
+			So(wC.Body.String(), ShouldEqual, fmt.Sprintf(`{"answer":%#q}`, sdp))
 		})
 	})
 }
@@ -480,12 +640,12 @@ func TestSnowflakeHeap(t *testing.T) {
 func TestInvalidGeoipFile(t *testing.T) {
 	Convey("Geoip", t, func() {
 		// Make sure things behave properly if geoip file fails to load
-		ctx := NewBrokerContext(NullLogger())
+		ctx := NewBrokerContext(NullLogger(), "", "")
 		if err := ctx.metrics.LoadGeoipDatabases("invalid_filename", "invalid_filename6"); err != nil {
 			log.Printf("loading geo ip databases returned error: %v", err)
 		}
 		ctx.metrics.UpdateCountryStats("127.0.0.1", "", NATUnrestricted)
-		So(ctx.metrics.geoipdb, ShouldEqual, nil)
+		So(ctx.metrics.geoipdb, ShouldBeNil)
 
 	})
 }
@@ -494,11 +654,11 @@ func TestMetrics(t *testing.T) {
 	Convey("Test metrics...", t, func() {
 		done := make(chan bool)
 		buf := new(bytes.Buffer)
-		ctx := NewBrokerContext(log.New(buf, "", 0))
+		ctx := NewBrokerContext(log.New(buf, "", 0), "", "")
 		i := &IPC{ctx}
 
 		err := ctx.metrics.LoadGeoipDatabases("test_geoip", "test_geoip6")
-		So(err, ShouldEqual, nil)
+		So(err, ShouldBeNil)
 
 		//Test addition of proxy polls
 		Convey("for proxy polls", func() {
@@ -560,21 +720,49 @@ func TestMetrics(t *testing.T) {
 			So(metricsStr, ShouldContainSubstring, "\nsnowflake-ips-standalone 1\n")
 			So(metricsStr, ShouldContainSubstring, "\nsnowflake-ips-badge 1\n")
 			So(metricsStr, ShouldContainSubstring, "\nsnowflake-ips-webext 1\n")
-			So(metricsStr, ShouldEndWith, "\nsnowflake-ips-total 4\nsnowflake-idle-count 8\nsnowflake-proxy-poll-with-relay-url-count 0\nsnowflake-proxy-poll-without-relay-url-count 8\nsnowflake-proxy-rejected-for-relay-url-count 0\nclient-denied-count 0\nclient-restricted-denied-count 0\nclient-unrestricted-denied-count 0\nclient-snowflake-match-count 0\nsnowflake-ips-nat-restricted 0\nsnowflake-ips-nat-unrestricted 0\nsnowflake-ips-nat-unknown 1\n")
+			So(metricsStr, ShouldEndWith, `snowflake-ips-total 4
+snowflake-idle-count 8
+snowflake-proxy-poll-with-relay-url-count 0
+snowflake-proxy-poll-without-relay-url-count 8
+snowflake-proxy-rejected-for-relay-url-count 0
+client-denied-count 0
+client-restricted-denied-count 0
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 0
+client-http-ips 
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+snowflake-ips-nat-restricted 0
+snowflake-ips-nat-unrestricted 0
+snowflake-ips-nat-unknown 1
+`)
 		})
 
 		//Test addition of client failures
 		Convey("for no proxies available", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unknown\"}"))
+			data, err := createClientOffer(sdp, NATUnknown, "")
+			So(err, ShouldBeNil)
 			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
+			r.RemoteAddr = "129.97.208.23:8888" //CA geoip
 			So(err, ShouldBeNil)
 
 			clientOffers(i, w, r)
 
 			ctx.metrics.printMetrics()
-			So(buf.String(), ShouldContainSubstring, "client-denied-count 8\nclient-restricted-denied-count 8\nclient-unrestricted-denied-count 0\nclient-snowflake-match-count 0")
+			So(buf.String(), ShouldContainSubstring, `client-denied-count 8
+client-restricted-denied-count 8
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 8
+client-http-ips CA=8
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips `)
 
 			// Test reset
 			buf.Reset()
@@ -584,13 +772,31 @@ func TestMetrics(t *testing.T) {
 			So(buf.String(), ShouldContainSubstring, "\nsnowflake-ips-standalone 0\n")
 			So(buf.String(), ShouldContainSubstring, "\nsnowflake-ips-badge 0\n")
 			So(buf.String(), ShouldContainSubstring, "\nsnowflake-ips-webext 0\n")
-			So(buf.String(), ShouldContainSubstring, "\nsnowflake-ips-total 0\nsnowflake-idle-count 0\nsnowflake-proxy-poll-with-relay-url-count 0\nsnowflake-proxy-poll-without-relay-url-count 0\nsnowflake-proxy-rejected-for-relay-url-count 0\nclient-denied-count 0\nclient-restricted-denied-count 0\nclient-unrestricted-denied-count 0\nclient-snowflake-match-count 0\nsnowflake-ips-nat-restricted 0\nsnowflake-ips-nat-unrestricted 0\nsnowflake-ips-nat-unknown 0\n")
+			So(buf.String(), ShouldContainSubstring, `snowflake-ips-total 0
+snowflake-idle-count 0
+snowflake-proxy-poll-with-relay-url-count 0
+snowflake-proxy-poll-without-relay-url-count 0
+snowflake-proxy-rejected-for-relay-url-count 0
+client-denied-count 0
+client-restricted-denied-count 0
+client-unrestricted-denied-count 0
+client-snowflake-match-count 0
+client-http-count 0
+client-http-ips 
+client-ampcache-count 0
+client-ampcache-ips 
+client-sqs-count 0
+client-sqs-ips 
+snowflake-ips-nat-restricted 0
+snowflake-ips-nat-unrestricted 0
+snowflake-ips-nat-unknown 0
+`)
 		})
 		//Test addition of client matches
 		Convey("for client-proxy match", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unknown\"}"))
+			data, err := createClientOffer(sdp, NATUnknown, "")
+			So(err, ShouldBeNil)
 			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 
@@ -601,7 +807,7 @@ func TestMetrics(t *testing.T) {
 				done <- true
 			}()
 			offer := <-snowflake.offerChannel
-			So(offer.sdp, ShouldResemble, []byte("fake"))
+			So(offer.sdp, ShouldResemble, []byte(sdp))
 			snowflake.answerChannel <- "fake answer"
 			<-done
 
@@ -611,51 +817,57 @@ func TestMetrics(t *testing.T) {
 		//Test rounding boundary
 		Convey("binning boundary", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err := createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
+			clientOffers(i, w, r)
 
-			clientOffers(i, w, r)
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
@@ -664,11 +876,12 @@ func TestMetrics(t *testing.T) {
 			So(buf.String(), ShouldContainSubstring, "client-denied-count 8\nclient-restricted-denied-count 8\nclient-unrestricted-denied-count 0\n")
 
 			w = httptest.NewRecorder()
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err = createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 			clientOffers(i, w, r)
+
 			buf.Reset()
 			ctx.metrics.printMetrics()
 			So(buf.String(), ShouldContainSubstring, "client-denied-count 16\nclient-restricted-denied-count 16\nclient-unrestricted-denied-count 0\n")
@@ -743,11 +956,11 @@ func TestMetrics(t *testing.T) {
 			ctx.metrics.printMetrics()
 			So(buf.String(), ShouldContainSubstring, "snowflake-ips-nat-restricted 1\nsnowflake-ips-nat-unrestricted 1\nsnowflake-ips-nat-unknown 0")
 		})
-		//Test client failures by NAT type
+
 		Convey("client failures by NAT type", func() {
 			w := httptest.NewRecorder()
-			data := bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"restricted\"}"))
+			data, err := createClientOffer(sdp, NATRestricted, "")
+			So(err, ShouldBeNil)
 			r, err := http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 
@@ -759,8 +972,8 @@ func TestMetrics(t *testing.T) {
 			buf.Reset()
 			ctx.metrics.zeroMetrics()
 
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unrestricted\"}"))
+			data, err = createClientOffer(sdp, NATUnrestricted, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 
@@ -772,8 +985,8 @@ func TestMetrics(t *testing.T) {
 			buf.Reset()
 			ctx.metrics.zeroMetrics()
 
-			data = bytes.NewReader(
-				[]byte("1.0\n{\"offer\": \"fake\", \"nat\": \"unknown\"}"))
+			data, err = createClientOffer(sdp, NATUnknown, "")
+			So(err, ShouldBeNil)
 			r, err = http.NewRequest("POST", "snowflake.broker/client", data)
 			So(err, ShouldBeNil)
 
