@@ -2,9 +2,10 @@ package event
 
 import (
 	"fmt"
+	"time"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/safelog"
 	"github.com/pion/webrtc/v3"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/ptutil/safelog"
 )
 
 type SnowflakeEvent interface {
@@ -66,6 +67,14 @@ func (e EventOnProxyStarting) String() string {
 	return "Proxy starting"
 }
 
+type EventOnProxyClientConnected struct {
+	SnowflakeEvent
+}
+
+func (e EventOnProxyClientConnected) String() string {
+	return fmt.Sprintf("client connected")
+}
+
 type EventOnProxyConnectionOver struct {
 	SnowflakeEvent
 	InboundTraffic  int64
@@ -73,7 +82,23 @@ type EventOnProxyConnectionOver struct {
 }
 
 func (e EventOnProxyConnectionOver) String() string {
-	return fmt.Sprintf("Proxy connection closed (↑ %d, ↓ %d)", e.InboundTraffic, e.OutboundTraffic)
+	return fmt.Sprintf("Proxy connection closed")
+}
+
+type EventOnProxyStats struct {
+	SnowflakeEvent
+	ConnectionCount             int
+	InboundBytes, OutboundBytes int64
+	InboundUnit, OutboundUnit   string
+	SummaryInterval             time.Duration
+}
+
+func (e EventOnProxyStats) String() string {
+	statString := fmt.Sprintf("In the last %v, there were %v completed connections. Traffic Relayed ↓ %v %v (%.2f %v%s), ↑ %v %v (%.2f %v%s).",
+		e.SummaryInterval.String(), e.ConnectionCount,
+		e.InboundBytes, e.InboundUnit, float64(e.InboundBytes)/e.SummaryInterval.Seconds(), e.InboundUnit, "/s",
+		e.OutboundBytes, e.OutboundUnit, float64(e.OutboundBytes)/e.SummaryInterval.Seconds(), e.OutboundUnit, "/s")
+	return statString
 }
 
 type EventOnCurrentNATTypeDetermined struct {
