@@ -1,15 +1,15 @@
 //Package for communication with the snowflake broker
 
-// import "git.torproject.org/pluggable-transports/snowflake.git/v2/common/messages"
+// import "gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/messages"
 package messages
 
 import (
 	"bytes"
 	"encoding/json"
 	"fmt"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/bridgefingerprint"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/bridgefingerprint"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/nat"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/nat"
 )
 
 const ClientVersion = "1.0"
