@@ -31,7 +31,6 @@ import (
 	"encoding/base64"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"log"
 	"net"
 	"net/http"
@@ -40,41 +39,54 @@ import (
 	"sync"
 	"time"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/event"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/messages"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/namematcher"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/task"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/util"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/websocketconn"
 	"github.com/gorilla/websocket"
 	"github.com/pion/ice/v2"
+	"github.com/pion/transport/v2/stdnet"
 	"github.com/pion/webrtc/v3"
+
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/event"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/messages"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/namematcher"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/task"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/util"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/websocketconn"
+
+	snowflakeClient "gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/client/lib"
 )
 
-const DefaultBrokerURL = "https://snowflake-broker.torproject.net/"
-
-const DefaultNATProbeURL = "https://snowflake-broker.torproject.net:8443/probe"
-
-const DefaultRelayURL = "wss://snowflake.bamsoftware.com/"
-
-const DefaultSTUNURL = "stun:stun.l.google.com:19302"
-const DefaultProxyType = "standalone"
-const pollInterval = 5 * time.Second
+const (
+	DefaultPollInterval = 5 * time.Second
+	DefaultBrokerURL    = "https://snowflake-broker.torproject.net/"
+	DefaultNATProbeURL  = "https://snowflake-broker.torproject.net:8443/probe"
+	// This is rather a "DefaultDefaultRelayURL"
+	DefaultRelayURL  = "wss://snowflake.torproject.net/"
+	DefaultSTUNURL   = "stun:stun.l.google.com:19302,stun:stun.voip.blackberry.com:3478"
+	DefaultProxyType = "standalone"
+)
 
 const (
-	// NATUnknown represents a NAT type which is unknown.
+	// NATUnknown is set if the proxy cannot connect to probetest.
 	NATUnknown = "unknown"
-	// NATRestricted represents a restricted NAT.
+
+	// NATRestricted is set if the proxy times out when connecting to a symmetric NAT.
 	NATRestricted = "restricted"
-	// NATUnrestricted represents an unrestricted NAT.
+
+	// NATUnrestricted is set if the proxy successfully connects to a symmetric NAT.
 	NATUnrestricted = "unrestricted"
 )
 
-// amount of time after sending an SDP answer before the proxy assumes the
-// client is not going to connect
-const dataChannelTimeout = 20 * time.Second
+const (
+	// Amount of time after sending an SDP answer before the proxy assumes the
+	// client is not going to connect
+	dataChannelTimeout = 20 * time.Second
 
-const readLimit = 100000 //Maximum number of bytes to be read from an HTTP request
+	// Maximum number of bytes to be read from an HTTP request
+	readLimit = 100000
+
+	sessionIDLength = 16
+)
+
+const bufferedAmountLowThreshold uint64 = 256 * 1024 // 256 KB
 
 var broker *SignalingServer
 
@@ -90,9 +102,11 @@ func getCurrentNATType() string {
 	return currentNATType
 }
 
-const (
-	sessionIDLength = 16
-)
+func setCurrentNATType(newType string) {
+	currentNATTypeAccess.Lock()
+	defer currentNATTypeAccess.Unlock()
+	currentNATType = newType
+}
 
 var (
 	tokens *tokens_t
@@ -102,19 +116,27 @@ var (
 
 // SnowflakeProxy is used to configure an embedded
 // Snowflake in another Go application.
+// For some more info also see CLI parameter descriptions in README.
 type SnowflakeProxy struct {
+	// How often to ask the broker for a new client
+	PollInterval time.Duration
 	// Capacity is the maximum number of clients a Snowflake will serve.
 	// Proxies with a capacity of 0 will accept an unlimited number of clients.
 	Capacity uint
-	// STUNURL is the URL of the STUN server the proxy will use
+	// STUNURL is the URLs (comma-separated) of the STUN server the proxy will use
 	STUNURL string
 	// BrokerURL is the URL of the Snowflake broker
 	BrokerURL string
 	// KeepLocalAddresses indicates whether local SDP candidates will be sent to the broker
 	KeepLocalAddresses bool
-	// RelayURL is the URL of the Snowflake server that all traffic will be relayed to
+	// RelayURL is the default `URL` of the server (relay)
+	// that this proxy will forward client connections to,
+	// in case the broker itself did not specify the said URL
 	RelayURL string
-	// Ephemeral*Port limits the pool of ports that ICE UDP connections can allocate from
+	// OutboundAddress specify an IP address to use as SDP host candidate
+	OutboundAddress string
+	// EphemeralMinPort and EphemeralMaxPort limit the range of ports that
+	// ICE UDP connections may allocate from.
 	EphemeralMinPort uint16
 	EphemeralMaxPort uint16
 	// RelayDomainNamePattern is the pattern specify allowed domain name for relay
@@ -123,7 +145,12 @@ type SnowflakeProxy struct {
 	// There is no look ahead assertion when matching domain name suffix,
 	// thus the string prepend the suffix does not need to be empty or ends with a dot.
 	RelayDomainNamePattern string
-	AllowNonTLSRelay       bool
+	// AllowProxyingToPrivateAddresses determines whether to allow forwarding
+	// client connections to private IP addresses.
+	// Useful when a Snowflake server (relay) is hosted on the same private network
+	// as this proxy.
+	AllowProxyingToPrivateAddresses bool
+	AllowNonTLSRelay                bool
 	// NATProbeURL is the URL of the probe service we use for NAT checks
 	NATProbeURL string
 	// NATTypeMeasurementInterval is time before NAT type is retested
@@ -132,6 +159,12 @@ type SnowflakeProxy struct {
 	ProxyType       string
 	EventDispatcher event.SnowflakeEventDispatcher
 	shutdown        chan struct{}
+
+	// SummaryInterval is the time interval at which proxy stats will be logged
+	SummaryInterval time.Duration
+
+	periodicProxyStats *periodicProxyStats
+	bytesLogger        bytesLogger
 }
 
 // Checks whether an IP address is a remote address for the client
@@ -149,7 +182,7 @@ func genSessionID() string {
 }
 
 func limitedRead(r io.Reader, limit int64) ([]byte, error) {
-	p, err := ioutil.ReadAll(&io.LimitedReader{R: r, N: limit + 1})
+	p, err := io.ReadAll(&io.LimitedReader{R: r, N: limit + 1})
 	if err != nil {
 		return p, err
 	} else if int64(len(p)) == limit+1 {
@@ -182,16 +215,15 @@ func newSignalingServer(rawURL string, keepLocalAddresses bool) (*SignalingServe
 
 // Post sends a POST request to the SignalingServer
 func (s *SignalingServer) Post(path string, payload io.Reader) ([]byte, error) {
-
 	req, err := http.NewRequest("POST", path, payload)
 	if err != nil {
 		return nil, err
 	}
+
 	resp, err := s.transport.RoundTrip(req)
 	if err != nil {
 		return nil, err
 	}
-
 	if resp.StatusCode != http.StatusOK {
 		return nil, fmt.Errorf("remote returned status code %d", resp.StatusCode)
 	}
@@ -200,52 +232,44 @@ func (s *SignalingServer) Post(path string, payload io.Reader) ([]byte, error) {
 	return limitedRead(resp.Body, readLimit)
 }
 
-func (s *SignalingServer) pollOffer(sid string, proxyType string, acceptedRelayPattern string, shutdown chan struct{}) (*webrtc.SessionDescription, string) {
+// pollOffer communicates the proxy's capabilities with broker
+// and retrieves a compatible SDP offer and relay URL.
+func (s *SignalingServer) pollOffer(sid string, proxyType string, acceptedRelayPattern string) (*webrtc.SessionDescription, string) {
 	brokerPath := s.url.ResolveReference(&url.URL{Path: "proxy"})
 
-	ticker := time.NewTicker(pollInterval)
-	defer ticker.Stop()
+	numClients := int((tokens.count() / 8) * 8) // Round down to 8
+	currentNATTypeLoaded := getCurrentNATType()
+	body, err := messages.EncodeProxyPollRequestWithRelayPrefix(sid, proxyType, currentNATTypeLoaded, numClients, acceptedRelayPattern)
+	if err != nil {
+		log.Printf("Error encoding poll message: %s", err.Error())
+		return nil, ""
+	}
 
-	// Run the loop once before hitting the ticker
-	for ; true; <-ticker.C {
-		select {
-		case <-shutdown:
+	resp, err := s.Post(brokerPath.String(), bytes.NewBuffer(body))
+	if err != nil {
+		log.Printf("error polling broker: %s", err.Error())
+	}
+
+	offer, _, relayURL, err := messages.DecodePollResponseWithRelayURL(resp)
+	if err != nil {
+		log.Printf("Error reading broker response: %s", err.Error())
+		log.Printf("body: %s", resp)
+		return nil, ""
+	}
+	if offer != "" {
+		offer, err := util.DeserializeSessionDescription(offer)
+		if err != nil {
+			log.Printf("Error processing session description: %s", err.Error())
 			return nil, ""
-		default:
-			numClients := int((tokens.count() / 8) * 8) // Round down to 8
-			currentNATTypeLoaded := getCurrentNATType()
-			body, err := messages.EncodeProxyPollRequestWithRelayPrefix(sid, proxyType, currentNATTypeLoaded, numClients, acceptedRelayPattern)
-			if err != nil {
-				log.Printf("Error encoding poll message: %s", err.Error())
-				return nil, ""
-			}
-			resp, err := s.Post(brokerPath.String(), bytes.NewBuffer(body))
-			if err != nil {
-				log.Printf("error polling broker: %s", err.Error())
-			}
-
-			offer, _, relayURL, err := messages.DecodePollResponseWithRelayURL(resp)
-			if err != nil {
-				log.Printf("Error reading broker response: %s", err.Error())
-				log.Printf("body: %s", resp)
-				return nil, ""
-			}
-			if offer != "" {
-				offer, err := util.DeserializeSessionDescription(offer)
-				if err != nil {
-					log.Printf("Error processing session description: %s", err.Error())
-					return nil, ""
-				}
-				return offer, relayURL
-
-			}
 		}
+		return offer, relayURL
 	}
 	return nil, ""
 }
 
+// sendAnswer encodes an SDP answer, sends it to the broker
+// and wait for its response
 func (s *SignalingServer) sendAnswer(sid string, pc *webrtc.PeerConnection) error {
-	brokerPath := s.url.ResolveReference(&url.URL{Path: "answer"})
 	ld := pc.LocalDescription()
 	if !s.keepLocalAddresses {
 		ld = &webrtc.SessionDescription{
@@ -253,14 +277,18 @@ func (s *SignalingServer) sendAnswer(sid string, pc *webrtc.PeerConnection) erro
 			SDP:  util.StripLocalAddresses(ld.SDP),
 		}
 	}
+
 	answer, err := util.SerializeSessionDescription(ld)
 	if err != nil {
 		return err
 	}
+
 	body, err := messages.EncodeAnswerRequest(answer, sid)
 	if err != nil {
 		return err
 	}
+
+	brokerPath := s.url.ResolveReference(&url.URL{Path: "answer"})
 	resp, err := s.Post(brokerPath.String(), bytes.NewBuffer(body))
 	if err != nil {
 		return fmt.Errorf("error sending answer to broker: %s", err.Error())
@@ -283,10 +311,15 @@ func copyLoop(c1 io.ReadWriteCloser, c2 io.ReadWriteCloser, shutdown chan struct
 	defer c1.Close()
 	done := make(chan struct{})
 	copyer := func(dst io.ReadWriteCloser, src io.ReadWriteCloser) {
+		// Experimentally each usage of buffer has been observed to be lower than
+		// 2K; io.Copy defaults to 32K.
+		// This is probably determined by MTU in the server's `newHTTPHandler`.
+		size := 2 * 1024
+		buffer := make([]byte, size)
 		// Ignore io.ErrClosedPipe because it is likely caused by the
 		// termination of copyer in the other direction.
-		if _, err := io.Copy(dst, src); err != nil && err != io.ErrClosedPipe {
-			log.Printf("io.Copy inside CopyLoop generated an error: %v", err)
+		if _, err := io.CopyBuffer(dst, src, buffer); err != nil && err != io.ErrClosedPipe {
+			log.Printf("io.CopyBuffer inside CopyLoop generated an error: %v", err)
 		}
 		once.Do(func() {
 			close(done)
@@ -305,8 +338,8 @@ func copyLoop(c1 io.ReadWriteCloser, c2 io.ReadWriteCloser, shutdown chan struct
 
 // We pass conn.RemoteAddr() as an additional parameter, rather than calling
 // conn.RemoteAddr() inside this function, as a workaround for a hang that
-// otherwise occurs inside of conn.pc.RemoteDescription() (called by
-// RemoteAddr). https://bugs.torproject.org/18628#comment:8
+// otherwise occurs inside conn.pc.RemoteDescription() (called by RemoteAddr).
+// https://bugs.torproject.org/18628#comment:8
 func (sf *SnowflakeProxy) datachannelHandler(conn *webRTCConn, remoteAddr net.Addr, relayURL string) {
 	defer conn.Close()
 	defer tokens.ret()
@@ -314,9 +347,11 @@ func (sf *SnowflakeProxy) datachannelHandler(conn *webRTCConn, remoteAddr net.Ad
 	if relayURL == "" {
 		relayURL = sf.RelayURL
 	}
+
 	u, err := url.Parse(relayURL)
 	if err != nil {
-		log.Fatalf("invalid relay url: %s", err)
+		log.Printf("invalid relay url: %s", err)
+		return
 	}
 
 	if remoteAddr != nil {
@@ -334,8 +369,9 @@ func (sf *SnowflakeProxy) datachannelHandler(conn *webRTCConn, remoteAddr net.Ad
 		log.Printf("error dialing relay: %s = %s", u.String(), err)
 		return
 	}
+
 	wsConn := websocketconn.New(ws)
-	log.Printf("connected to relay: %v", relayURL)
+	log.Printf("Connected to relay: %v", relayURL)
 	defer wsConn.Close()
 	copyLoop(conn, wsConn, sf.shutdown)
 	log.Printf("datachannelHandler ends")
@@ -353,6 +389,13 @@ func (d dataChannelHandlerWithRelayURL) datachannelHandler(conn *webRTCConn, rem
 func (sf *SnowflakeProxy) makeWebRTCAPI() *webrtc.API {
 	settingsEngine := webrtc.SettingEngine{}
 
+	// Use the SetNet setting https://pkg.go.dev/github.com/pion/webrtc/v3#SettingEngine.SetNet
+	// to get snowflake working in shadow (where the AF_NETLINK family is not implemented).
+	// These two lines of code functionally revert a new change in pion by silently ignoring
+	// when net.Interfaces() fails, rather than throwing an error
+	vnet, _ := stdnet.NewNet()
+	settingsEngine.SetNet(vnet)
+
 	if sf.EphemeralMinPort != 0 && sf.EphemeralMaxPort != 0 {
 		err := settingsEngine.SetEphemeralUDPPortRange(sf.EphemeralMinPort, sf.EphemeralMaxPort)
 		if err != nil {
@@ -360,8 +403,16 @@ func (sf *SnowflakeProxy) makeWebRTCAPI() *webrtc.API {
 		}
 	}
 
+	if sf.OutboundAddress != "" {
+		// replace SDP host candidates with the given IP without validation
+		// still have server reflexive candidates to fall back on
+		settingsEngine.SetNAT1To1IPs([]string{sf.OutboundAddress}, webrtc.ICECandidateTypeHost)
+	}
+
 	settingsEngine.SetICEMulticastDNSMode(ice.MulticastDNSModeDisabled)
 
+	settingsEngine.SetDTLSInsecureSkipHelloVerify(true)
+
 	return webrtc.NewAPI(webrtc.WithSettingEngine(settingsEngine))
 }
 
@@ -369,36 +420,54 @@ func (sf *SnowflakeProxy) makeWebRTCAPI() *webrtc.API {
 // candidates is complete and the answer is available in LocalDescription.
 // Installs an OnDataChannel callback that creates a webRTCConn and passes it to
 // datachannelHandler.
-func (sf *SnowflakeProxy) makePeerConnectionFromOffer(sdp *webrtc.SessionDescription,
-	config webrtc.Configuration,
-	dataChan chan struct{},
-	handler func(conn *webRTCConn, remoteAddr net.Addr)) (*webrtc.PeerConnection, error) {
-
+func (sf *SnowflakeProxy) makePeerConnectionFromOffer(
+	sdp *webrtc.SessionDescription,
+	config webrtc.Configuration, dataChan chan struct{},
+	handler func(conn *webRTCConn, remoteAddr net.Addr),
+) (*webrtc.PeerConnection, error) {
 	api := sf.makeWebRTCAPI()
 	pc, err := api.NewPeerConnection(config)
 	if err != nil {
 		return nil, fmt.Errorf("accept: NewPeerConnection: %s", err)
 	}
+
 	pc.OnDataChannel(func(dc *webrtc.DataChannel) {
-		log.Println("OnDataChannel")
+		log.Printf("New Data Channel %s-%d\n", dc.Label(), dc.ID())
 		close(dataChan)
 
 		pr, pw := io.Pipe()
-		conn := newWebRTCConn(pc, dc, pr, sf.EventDispatcher)
+		conn := newWebRTCConn(pc, dc, pr, sf.bytesLogger)
+
+		dc.SetBufferedAmountLowThreshold(bufferedAmountLowThreshold)
+
+		dc.OnBufferedAmountLow(func() {
+			select {
+			case conn.sendMoreCh <- struct{}{}:
+			default:
+			}
+		})
 
 		dc.OnOpen(func() {
-			log.Println("OnOpen channel")
+			log.Printf("Data Channel %s-%d open\n", dc.Label(), dc.ID())
+			sf.EventDispatcher.OnNewSnowflakeEvent(event.EventOnProxyClientConnected{})
+
+			if sf.OutboundAddress != "" {
+				selectedCandidatePair, err := pc.SCTP().Transport().ICETransport().GetSelectedCandidatePair()
+				if err != nil {
+					log.Printf("Warning: couldn't get the selected candidate pair")
+				}
+
+				log.Printf("Selected Local Candidate: %s:%d", selectedCandidatePair.Local.Address, selectedCandidatePair.Local.Port)
+				if sf.OutboundAddress != selectedCandidatePair.Local.Address {
+					log.Printf("Warning: the IP address provided by --outbound-address is not used for establishing peerconnection")
+				}
+			}
 		})
 		dc.OnClose(func() {
 			conn.lock.Lock()
 			defer conn.lock.Unlock()
-			log.Println("OnClose channel")
-			log.Println(conn.bytesLogger.ThroughputSummary())
-			in, out := conn.bytesLogger.GetStat()
-			conn.eventLogger.OnNewSnowflakeEvent(event.EventOnProxyConnectionOver{
-				InboundTraffic:  in,
-				OutboundTraffic: out,
-			})
+			log.Printf("Data Channel %s-%d close\n", dc.Label(), dc.ID())
+			sf.EventDispatcher.OnNewSnowflakeEvent(event.EventOnProxyConnectionOver{})
 			conn.dc = nil
 			dc.Close()
 			pw.Close()
@@ -407,12 +476,17 @@ func (sf *SnowflakeProxy) makePeerConnectionFromOffer(sdp *webrtc.SessionDescrip
 			var n int
 			n, err = pw.Write(msg.Data)
 			if err != nil {
-				if inerr := pw.CloseWithError(err); inerr != nil {
-					log.Printf("close with error generated an error: %v", inerr)
+				if inErr := pw.CloseWithError(err); inErr != nil {
+					log.Printf("close with error generated an error: %v", inErr)
 				}
+
+				return
 			}
+
 			conn.bytesLogger.AddOutbound(int64(n))
+
 			if n != len(msg.Data) {
+				// XXX: Maybe don't panic here and log an error instead?
 				panic("short write")
 			}
 		})
@@ -430,7 +504,6 @@ func (sf *SnowflakeProxy) makePeerConnectionFromOffer(sdp *webrtc.SessionDescrip
 		}
 		return nil, fmt.Errorf("accept: SetRemoteDescription: %s", err)
 	}
-	log.Println("sdp offer successfully received.")
 
 	log.Println("Generating answer...")
 	answer, err := pc.CreateAnswer(nil)
@@ -451,24 +524,38 @@ func (sf *SnowflakeProxy) makePeerConnectionFromOffer(sdp *webrtc.SessionDescrip
 		}
 		return nil, err
 	}
-	// Wait for ICE candidate gathering to complete
-	<-done
+
+	// Wait for ICE candidate gathering to complete,
+	// or for whatever we managed to gather before the client times out.
+	// See https://gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/-/issues/40230
+	select {
+	case <-done:
+	case <-time.After(snowflakeClient.DataChannelTimeout / 2):
+		log.Print("ICE gathering is not yet complete, but let's send the answer" +
+			" before the client times out")
+	}
+
+	log.Printf("Answer: \n\t%s", strings.ReplaceAll(pc.LocalDescription().SDP, "\n", "\n\t"))
+
 	return pc, nil
 }
 
 // Create a new PeerConnection. Blocks until the gathering of ICE
 // candidates is complete and the answer is available in LocalDescription.
-func (sf *SnowflakeProxy) makeNewPeerConnection(config webrtc.Configuration,
-	dataChan chan struct{}) (*webrtc.PeerConnection, error) {
-
+func (sf *SnowflakeProxy) makeNewPeerConnection(
+	config webrtc.Configuration, dataChan chan struct{},
+) (*webrtc.PeerConnection, error) {
 	api := sf.makeWebRTCAPI()
 	pc, err := api.NewPeerConnection(config)
 	if err != nil {
 		return nil, fmt.Errorf("accept: NewPeerConnection: %s", err)
 	}
+	pc.OnConnectionStateChange(func(pcs webrtc.PeerConnectionState) {
+		log.Printf("NAT check: WebRTC: OnConnectionStateChange: %v", pcs)
+	})
 
 	// Must create a data channel before creating an offer
-	// https://github.com/pion/webrtc/wiki/Release-WebRTC@v3.0.0
+	// https://github.com/pion/webrtc/wiki/Release-WebRTC@v3.0.0#a-data-channel-is-no-longer-implicitly-created-with-a-peerconnection
 	dc, err := pc.CreateDataChannel("test", &webrtc.DataChannelInit{})
 	if err != nil {
 		log.Printf("CreateDataChannel ERROR: %s", err)
@@ -490,44 +577,44 @@ func (sf *SnowflakeProxy) makeNewPeerConnection(config webrtc.Configuration,
 		pc.Close()
 		return nil, err
 	}
-	log.Println("WebRTC: Created offer")
+	log.Println("Probetest: Created Offer")
 
 	// As of v3.0.0, pion-webrtc uses trickle ICE by default.
 	// We have to wait for candidate gathering to complete
 	// before we send the offer
 	done := webrtc.GatheringCompletePromise(pc)
+	// start the gathering of ICE candidates
 	err = pc.SetLocalDescription(offer)
 	if err != nil {
-		log.Println("Failed to prepare offer", err)
+		log.Println("Failed to apply offer", err)
 		pc.Close()
 		return nil, err
 	}
-	log.Println("WebRTC: Set local description")
+	log.Println("Probetest: Set local description")
 
 	// Wait for ICE candidate gathering to complete
 	<-done
+
 	return pc, nil
 }
 
 func (sf *SnowflakeProxy) runSession(sid string) {
-	offer, relayURL := broker.pollOffer(sid, sf.ProxyType, sf.RelayDomainNamePattern, sf.shutdown)
+	offer, relayURL := broker.pollOffer(sid, sf.ProxyType, sf.RelayDomainNamePattern)
 	if offer == nil {
 		log.Printf("bad offer from broker")
 		tokens.ret()
 		return
 	}
-	matcher := namematcher.NewNameMatcher(sf.RelayDomainNamePattern)
-	parsedRelayURL, err := url.Parse(relayURL)
-	if err != nil {
-		log.Printf("bad offer from broker: bad Relay URL %v", err.Error())
-		tokens.ret()
-		return
-	}
-	if relayURL != "" && (!matcher.IsMember(parsedRelayURL.Hostname()) || (!sf.AllowNonTLSRelay && parsedRelayURL.Scheme != "wss")) {
-		log.Printf("bad offer from broker: rejected Relay URL")
-		tokens.ret()
-		return
+	log.Printf("Received Offer From Broker: \n\t%s", strings.ReplaceAll(offer.SDP, "\n", "\n\t"))
+
+	if relayURL != "" {
+		if err := checkIsRelayURLAcceptable(sf.RelayDomainNamePattern, sf.AllowProxyingToPrivateAddresses, sf.AllowNonTLSRelay, relayURL); err != nil {
+			log.Printf("bad offer from broker: %v", err)
+			tokens.ret()
+			return
+		}
 	}
+
 	dataChan := make(chan struct{})
 	dataChannelAdaptor := dataChannelHandlerWithRelayURL{RelayURL: relayURL, sf: sf}
 	pc, err := sf.makePeerConnectionFromOffer(offer, config, dataChan, dataChannelAdaptor.datachannelHandler)
@@ -536,6 +623,7 @@ func (sf *SnowflakeProxy) runSession(sid string) {
 		tokens.ret()
 		return
 	}
+
 	err = broker.sendAnswer(sid, pc)
 	if err != nil {
 		log.Printf("error sending answer to client through broker: %s", err)
@@ -550,7 +638,7 @@ func (sf *SnowflakeProxy) runSession(sid string) {
 	// destroy the peer connection and return the token.
 	select {
 	case <-dataChan:
-		log.Println("Connection successful.")
+		log.Println("Connection successful")
 	case <-time.After(dataChannelTimeout):
 		log.Println("Timed out waiting for client to open data channel.")
 		if err := pc.Close(); err != nil {
@@ -560,6 +648,42 @@ func (sf *SnowflakeProxy) runSession(sid string) {
 	}
 }
 
+// Returns nil if the relayURL is acceptable
+func checkIsRelayURLAcceptable(
+	allowedHostNamePattern string,
+	allowPrivateIPs bool,
+	allowNonTLSRelay bool,
+	relayURL string,
+) error {
+	parsedRelayURL, err := url.Parse(relayURL)
+	if err != nil {
+		return fmt.Errorf("bad Relay URL %w", err)
+	}
+	if !allowPrivateIPs {
+		ip := net.ParseIP(parsedRelayURL.Hostname())
+		// Otherwise it's a domain name, or an invalid IP.
+		if ip != nil {
+			// We should probably use a ready library for this.
+			if !isRemoteAddress(ip) {
+				return fmt.Errorf("rejected Relay URL: private IPs are not allowed")
+			}
+		}
+	}
+	if !allowNonTLSRelay && parsedRelayURL.Scheme != "wss" {
+		return fmt.Errorf("rejected Relay URL protocol: non-TLS not allowed")
+	}
+	// FYI our websocket library also rejects other protocols
+	// https://github.com/gorilla/websocket/blob/5e002381133d322c5f1305d171f3bdd07decf229/client.go#L174-L181
+	if parsedRelayURL.Scheme != "wss" && parsedRelayURL.Scheme != "ws" {
+		return fmt.Errorf("rejected Relay URL protocol: only WebSocket is allowed")
+	}
+	matcher := namematcher.NewNameMatcher(allowedHostNamePattern)
+	if !matcher.IsMember(parsedRelayURL.Hostname()) {
+		return fmt.Errorf("rejected Relay URL: hostname does not match allowed pattern \"%v\"", allowedHostNamePattern)
+	}
+	return nil
+}
+
 // Start configures and starts a Snowflake, fully formed and special. Configuration
 // values that are unset will default to their corresponding default values.
 func (sf *SnowflakeProxy) Start() error {
@@ -569,6 +693,9 @@ func (sf *SnowflakeProxy) Start() error {
 	sf.shutdown = make(chan struct{})
 
 	// blank configurations revert to default
+	if sf.PollInterval == 0 {
+		sf.PollInterval = DefaultPollInterval
+	}
 	if sf.BrokerURL == "" {
 		sf.BrokerURL = DefaultBrokerURL
 	}
@@ -588,6 +715,10 @@ func (sf *SnowflakeProxy) Start() error {
 		sf.EventDispatcher = event.NewSnowflakeEventDispatcher()
 	}
 
+	sf.bytesLogger = newBytesSyncLogger()
+	sf.periodicProxyStats = newPeriodicProxyStats(sf.SummaryInterval, sf.EventDispatcher, sf.bytesLogger)
+	sf.EventDispatcher.AddSnowflakeEventListener(sf.periodicProxyStats)
+
 	broker, err = newSignalingServer(sf.BrokerURL, sf.KeepLocalAddresses)
 	if err != nil {
 		return fmt.Errorf("error configuring broker: %s", err)
@@ -599,7 +730,7 @@ func (sf *SnowflakeProxy) Start() error {
 	}
 	_, err = url.Parse(sf.RelayURL)
 	if err != nil {
-		return fmt.Errorf("invalid relay url: %s", err)
+		return fmt.Errorf("invalid default relay url: %s", err)
 	}
 
 	if !namematcher.IsValidRule(sf.RelayDomainNamePattern) {
@@ -609,24 +740,28 @@ func (sf *SnowflakeProxy) Start() error {
 	config = webrtc.Configuration{
 		ICEServers: []webrtc.ICEServer{
 			{
-				URLs: []string{sf.STUNURL},
+				URLs: strings.Split(sf.STUNURL, ","),
 			},
 		},
 	}
 	tokens = newTokens(sf.Capacity)
 
-	// use probetest to determine NAT compatability
-	sf.checkNATType(config, sf.NATProbeURL)
-
-	currentNATTypeLoaded := getCurrentNATType()
-
-	sf.EventDispatcher.OnNewSnowflakeEvent(&event.EventOnCurrentNATTypeDetermined{CurNATType: currentNATTypeLoaded})
+	err = sf.checkNATType(config, sf.NATProbeURL)
+	if err != nil {
+		// non-fatal error. Log it and continue
+		log.Printf(err.Error())
+		setCurrentNATType(NATUnknown)
+	}
+	sf.EventDispatcher.OnNewSnowflakeEvent(&event.EventOnCurrentNATTypeDetermined{CurNATType: getCurrentNATType()})
 
 	NatRetestTask := task.Periodic{
 		Interval: sf.NATTypeMeasurementInterval,
 		Execute: func() error {
-			sf.checkNATType(config, sf.NATProbeURL)
-			return nil
+			return sf.checkNATType(config, sf.NATProbeURL)
+		},
+		// Not setting OnError would shut down the periodic task on error by default.
+		OnError: func(err error) {
+			log.Printf("Periodic probetest failed: %s, retaining current NAT type: %s", err.Error(), getCurrentNATType())
 		},
 	}
 
@@ -635,7 +770,7 @@ func (sf *SnowflakeProxy) Start() error {
 		defer NatRetestTask.Close()
 	}
 
-	ticker := time.NewTicker(pollInterval)
+	ticker := time.NewTicker(sf.PollInterval)
 	defer ticker.Stop()
 
 	for ; true; <-ticker.C {
@@ -656,87 +791,83 @@ func (sf *SnowflakeProxy) Stop() {
 	close(sf.shutdown)
 }
 
-func (sf *SnowflakeProxy) checkNATType(config webrtc.Configuration, probeURL string) {
+// checkNATType use probetest to determine NAT compatability by
+// attempting to connect with a known symmetric NAT. If success,
+// it is considered "unrestricted". If timeout it is considered "restricted"
+func (sf *SnowflakeProxy) checkNATType(config webrtc.Configuration, probeURL string) error {
+	log.Printf("Checking our NAT type, contacting NAT check probe server at \"%v\"...", probeURL)
 
 	probe, err := newSignalingServer(probeURL, false)
 	if err != nil {
-		log.Printf("Error parsing url: %s", err.Error())
+		return fmt.Errorf("Error parsing url: %w", err)
 	}
 
-	// create offer
 	dataChan := make(chan struct{})
 	pc, err := sf.makeNewPeerConnection(config, dataChan)
 	if err != nil {
-		log.Printf("error making WebRTC connection: %s", err)
-		return
+		return fmt.Errorf("Error making WebRTC connection: %w", err)
 	}
+	defer func() {
+		if err := pc.Close(); err != nil {
+			log.Printf("Probetest: error calling pc.Close: %v", err)
+		}
+	}()
 
 	offer := pc.LocalDescription()
+	log.Printf("Probetest offer: \n\t%s", strings.ReplaceAll(offer.SDP, "\n", "\n\t"))
 	sdp, err := util.SerializeSessionDescription(offer)
-	log.Printf("Offer: %s", sdp)
 	if err != nil {
-		log.Printf("Error encoding probe message: %s", err.Error())
-		return
+		return fmt.Errorf("Error encoding probe message: %w", err)
 	}
 
 	// send offer
 	body, err := messages.EncodePollResponse(sdp, true, "")
 	if err != nil {
-		log.Printf("Error encoding probe message: %s", err.Error())
-		return
+		return fmt.Errorf("Error encoding probe message: %w", err)
 	}
+
 	resp, err := probe.Post(probe.url.String(), bytes.NewBuffer(body))
 	if err != nil {
-		log.Printf("error polling probe: %s", err.Error())
-		return
+		return fmt.Errorf("Error polling probe: %w", err)
 	}
 
 	sdp, _, err = messages.DecodeAnswerRequest(resp)
 	if err != nil {
-		log.Printf("Error reading probe response: %s", err.Error())
-		return
+		return fmt.Errorf("Error reading probe response: %w", err)
 	}
+
 	answer, err := util.DeserializeSessionDescription(sdp)
 	if err != nil {
-		log.Printf("Error setting answer: %s", err.Error())
-		return
+		return fmt.Errorf("Error setting answer: %w", err)
 	}
+	log.Printf("Probetest answer: \n\t%s", strings.ReplaceAll(answer.SDP, "\n", "\n\t"))
+
 	err = pc.SetRemoteDescription(*answer)
 	if err != nil {
-		log.Printf("Error setting answer: %s", err.Error())
-		return
+		return fmt.Errorf("Error setting answer: %w", err)
 	}
 
-	currentNATTypeLoaded := getCurrentNATType()
+	prevNATType := getCurrentNATType()
 
-	currentNATTypeTestResult := NATUnknown
+	log.Printf("Waiting for a test WebRTC connection with NAT check probe server to establish...")
 	select {
 	case <-dataChan:
-		currentNATTypeTestResult = NATUnrestricted
+		log.Printf(
+			"Test WebRTC connection with NAT check probe server established!"+
+				" This means our NAT is %v!",
+			NATUnrestricted,
+		)
+		setCurrentNATType(NATUnrestricted)
 	case <-time.After(dataChannelTimeout):
-		currentNATTypeTestResult = NATRestricted
+		log.Printf(
+			"Test WebRTC connection with NAT check probe server timed out."+
+				" This means our NAT is %v.",
+			NATRestricted,
+		)
+		setCurrentNATType(NATRestricted)
 	}
 
-	currentNATTypeToStore := NATUnknown
-	switch currentNATTypeLoaded + "->" + currentNATTypeTestResult {
-	case NATUnrestricted + "->" + NATUnknown:
-		currentNATTypeToStore = NATUnrestricted
-
-	case NATRestricted + "->" + NATUnknown:
-		currentNATTypeToStore = NATRestricted
-
-	default:
-		currentNATTypeToStore = currentNATTypeTestResult
-	}
-
-	log.Printf("NAT Type measurement: %v -> %v = %v\n", currentNATTypeLoaded, currentNATTypeTestResult, currentNATTypeToStore)
-
-	currentNATTypeAccess.Lock()
-	currentNATType = currentNATTypeToStore
-	currentNATTypeAccess.Unlock()
-
-	if err := pc.Close(); err != nil {
-		log.Printf("error calling pc.Close: %v", err)
-	}
+	log.Printf("NAT Type measurement: %v -> %v\n", prevNATType, getCurrentNATType())
 
+	return nil
 }
