@@ -1,7 +1,6 @@
 package snowflake_proxy
 
 import (
-	"fmt"
 	"time"
 )
 
@@ -10,7 +9,6 @@ import (
 type bytesLogger interface {
 	AddOutbound(int64)
 	AddInbound(int64)
-	ThroughputSummary() string
 	GetStat() (in int64, out int64)
 }
 
@@ -23,25 +21,28 @@ func (b bytesNullLogger) AddOutbound(amount int64) {}
 // AddInbound in bytesNullLogger does nothing
 func (b bytesNullLogger) AddInbound(amount int64) {}
 
-// ThroughputSummary in bytesNullLogger does nothing
-func (b bytesNullLogger) ThroughputSummary() string { return "" }
-
 func (b bytesNullLogger) GetStat() (in int64, out int64) { return -1, -1 }
 
 // bytesSyncLogger uses channels to safely log from multiple sources with output
 // occuring at reasonable intervals.
 type bytesSyncLogger struct {
 	outboundChan, inboundChan chan int64
-	outbound, inbound         int64
+	statsChan                 chan bytesLoggerStats
+	stats                     bytesLoggerStats
 	outEvents, inEvents       int
 	start                     time.Time
 }
 
+type bytesLoggerStats struct {
+	outbound, inbound int64
+}
+
 // newBytesSyncLogger returns a new bytesSyncLogger and starts it loggin.
 func newBytesSyncLogger() *bytesSyncLogger {
 	b := &bytesSyncLogger{
 		outboundChan: make(chan int64, 5),
 		inboundChan:  make(chan int64, 5),
+		statsChan:    make(chan bytesLoggerStats),
 	}
 	go b.log()
 	b.start = time.Now()
@@ -52,11 +53,16 @@ func (b *bytesSyncLogger) log() {
 	for {
 		select {
 		case amount := <-b.outboundChan:
-			b.outbound += amount
+			b.stats.outbound += amount
 			b.outEvents++
 		case amount := <-b.inboundChan:
-			b.inbound += amount
+			b.stats.inbound += amount
 			b.inEvents++
+		case b.statsChan <- b.stats:
+			b.stats.inbound = 0
+			b.stats.outbound = 0
+			b.inEvents = 0
+			b.outEvents = 0
 		}
 	}
 }
@@ -71,18 +77,10 @@ func (b *bytesSyncLogger) AddInbound(amount int64) {
 	b.inboundChan <- amount
 }
 
-// ThroughputSummary view a formatted summary of the throughput totals
-func (b *bytesSyncLogger) ThroughputSummary() string {
-	inbound := b.inbound
-	outbound := b.outbound
-
-	inbound, inUnit := formatTraffic(inbound)
-	outbound, outUnit := formatTraffic(outbound)
-
-	t := time.Now()
-	return fmt.Sprintf("Traffic throughput (up|down): %d %s|%d %s -- (%d OnMessages, %d Sends, over %d seconds)", inbound, inUnit, outbound, outUnit, b.outEvents, b.inEvents, int(t.Sub(b.start).Seconds()))
+// GetStat returns the current inbound and outbound stats from the logger and then zeros the counts
+func (b *bytesSyncLogger) GetStat() (in int64, out int64) {
+	stats := <-b.statsChan
+	return stats.inbound, stats.outbound
 }
 
-func (b *bytesSyncLogger) GetStat() (in int64, out int64) { return b.inbound, b.outbound }
-
 func formatTraffic(amount int64) (value int64, unit string) { return amount / 1000, "KB" }
