@@ -2,6 +2,7 @@ package snowflake_proxy
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"io"
 	"log"
@@ -10,12 +11,13 @@ import (
 	"sync"
 	"time"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/event"
 	"github.com/pion/ice/v2"
 	"github.com/pion/sdp/v3"
 	"github.com/pion/webrtc/v3"
 )
 
+const maxBufferedAmount uint64 = 512 * 1024 // 512 KB
+
 var remoteIPPatterns = []*regexp.Regexp{
 	/* IPv4 */
 	regexp.MustCompile(`(?m)^c=IN IP4 ([\d.]+)(?:(?:\/\d+)?\/\d+)?(:? |\r?\n)`),
@@ -31,18 +33,21 @@ type webRTCConn struct {
 	lock sync.Mutex // Synchronization for DataChannel destruction
 	once sync.Once  // Synchronization for PeerConnection destruction
 
-	bytesLogger bytesLogger
-	eventLogger event.SnowflakeEventReceiver
+	isClosing bool
 
 	inactivityTimeout time.Duration
 	activity          chan struct{}
+	sendMoreCh        chan struct{}
 	cancelTimeoutLoop context.CancelFunc
+
+	bytesLogger bytesLogger
 }
 
-func newWebRTCConn(pc *webrtc.PeerConnection, dc *webrtc.DataChannel, pr *io.PipeReader, eventLogger event.SnowflakeEventReceiver) *webRTCConn {
-	conn := &webRTCConn{pc: pc, dc: dc, pr: pr, eventLogger: eventLogger}
-	conn.bytesLogger = newBytesSyncLogger()
+func newWebRTCConn(pc *webrtc.PeerConnection, dc *webrtc.DataChannel, pr *io.PipeReader, bytesLogger bytesLogger) *webRTCConn {
+	conn := &webRTCConn{pc: pc, dc: dc, pr: pr, bytesLogger: bytesLogger}
+	conn.isClosing = false
 	conn.activity = make(chan struct{}, 100)
+	conn.sendMoreCh = make(chan struct{}, 1)
 	conn.inactivityTimeout = 30 * time.Second
 	ctx, cancel := context.WithCancel(context.Background())
 	conn.cancelTimeoutLoop = cancel
@@ -55,7 +60,7 @@ func (c *webRTCConn) timeoutLoop(ctx context.Context) {
 	for {
 		select {
 		case <-timer.C:
-			c.Close()
+			_ = c.Close()
 			log.Println("Closed connection due to inactivity")
 			return
 		case <-c.activity:
@@ -76,19 +81,30 @@ func (c *webRTCConn) Read(b []byte) (int, error) {
 
 func (c *webRTCConn) Write(b []byte) (int, error) {
 	c.bytesLogger.AddInbound(int64(len(b)))
-	c.activity <- struct{}{}
+	select {
+	case c.activity <- struct{}{}:
+	default:
+	}
 	c.lock.Lock()
 	defer c.lock.Unlock()
 	if c.dc != nil {
-		c.dc.Send(b)
+		_ = c.dc.Send(b)
+		if !c.isClosing && c.dc.BufferedAmount() >= maxBufferedAmount {
+			<-c.sendMoreCh
+		}
 	}
 	return len(b), nil
 }
 
 func (c *webRTCConn) Close() (err error) {
+	c.isClosing = true
+	select {
+	case c.sendMoreCh <- struct{}{}:
+	default:
+	}
 	c.once.Do(func() {
 		c.cancelTimeoutLoop()
-		err = c.pc.Close()
+		err = errors.Join(c.pr.Close(), c.pc.Close())
 	})
 	return
 }
