@@ -44,17 +44,18 @@ import (
 	"sync"
 	"time"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/turbotunnel"
 	"github.com/xtaci/kcp-go/v5"
 	"github.com/xtaci/smux"
 	"golang.org/x/net/http2"
+
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/turbotunnel"
 )
 
 const (
 	// WindowSize is the number of packets in the send and receive window of a KCP connection.
 	WindowSize = 65535
 	// StreamSize controls the maximum amount of in flight data between a client and server.
-	StreamSize = 1048576 //1MB
+	StreamSize = 1048576 // 1MB
 )
 
 // Transport is a structure with methods that conform to the Go PT v2.1 API
@@ -65,7 +66,6 @@ type Transport struct {
 
 // NewSnowflakeServer returns a new server-side Transport for Snowflake.
 func NewSnowflakeServer(getCertificate func(*tls.ClientHelloInfo) (*tls.Certificate, error)) *Transport {
-
 	return &Transport{getCertificate: getCertificate}
 }
 
@@ -79,7 +79,11 @@ func (t *Transport) Listen(addr net.Addr, numKCPInstances int) (*SnowflakeListen
 		ln:     make([]*kcp.Listener, 0, numKCPInstances),
 	}
 
-	handler := newHTTPHandler(addr, numKCPInstances)
+	// kcp-go doesn't provide an accessor for the current MTU setting (and
+	// anyway we could not create a kcp.Listener without creating a
+	// net.PacketConn for it first), so assume the default kcp.IKCP_MTU_DEF
+	// (1400 bytes) and don't increase it elsewhere.
+	handler := newHTTPHandler(addr, numKCPInstances, kcp.IKCP_MTU_DEF)
 	server := &http.Server{
 		Addr:        addr.String(),
 		Handler:     handler,
@@ -125,13 +129,15 @@ func (t *Transport) Listen(addr net.Addr, numKCPInstances int) (*SnowflakeListen
 			errChan <- err
 		}
 	}()
-
 	select {
 	case err = <-errChan:
 		break
 	case <-time.After(listenAndServeErrorTimeout):
 		break
 	}
+	if err != nil {
+		return nil, err
+	}
 
 	listener.server = server
 
@@ -157,7 +163,6 @@ func (t *Transport) Listen(addr net.Addr, numKCPInstances int) (*SnowflakeListen
 	}
 
 	return listener, nil
-
 }
 
 type SnowflakeListener struct {
@@ -175,7 +180,7 @@ type SnowflakeListener struct {
 func (l *SnowflakeListener) Accept() (net.Conn, error) {
 	select {
 	case <-l.closed:
-		//channel has been closed, no longer accepting connections
+		// channel has been closed, no longer accepting connections
 		return nil, io.ErrClosedPipe
 	case conn := <-l.queue:
 		return conn, nil
@@ -297,6 +302,7 @@ func (conn *SnowflakeClientConn) SetDeadline(t time.Time) error { return conn.st
 func (conn *SnowflakeClientConn) SetReadDeadline(t time.Time) error {
 	return conn.stream.SetReadDeadline(t)
 }
+
 func (conn *SnowflakeClientConn) SetWriteDeadline(t time.Time) error {
 	return conn.stream.SetWriteDeadline(t)
 }
