@@ -7,7 +7,6 @@ import (
 	"flag"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"log"
 	"net"
 	"net/http"
@@ -19,12 +18,12 @@ import (
 	"sync"
 	"syscall"
 
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/safelog"
-	"git.torproject.org/pluggable-transports/snowflake.git/v2/common/version"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/ptutil/safelog"
+	"gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/common/version"
 	"golang.org/x/crypto/acme/autocert"
 
-	pt "git.torproject.org/pluggable-transports/goptlib.git"
-	sf "git.torproject.org/pluggable-transports/snowflake.git/v2/server/lib"
+	pt "gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/goptlib"
+	sf "gitlab.torproject.org/tpo/anti-censorship/pluggable-transports/snowflake/v2/server/lib"
 )
 
 const ptMethodName = "snowflake"
@@ -76,7 +75,9 @@ func handleConn(conn net.Conn, orPortSrcAddr *net.IPNet) error {
 	addr := conn.RemoteAddr().String()
 	statsChannel <- addr != ""
 
-	dialer := net.Dialer{}
+	dialer := net.Dialer{
+		Control: dialerControl,
+	}
 	if orPortSrcAddr != nil {
 		// Use a random source IP address in the given range.
 		ip, err := randIPAddr(orPortSrcAddr)
@@ -179,6 +180,7 @@ func main() {
 	if err != nil {
 		log.Fatalf("error in setup: %s", err)
 	}
+	pt.ReportVersion("snowflake-server", version.GetVersion())
 
 	go statsThread()
 
@@ -318,8 +320,8 @@ func main() {
 		// This environment variable means we should treat EOF on stdin
 		// just like SIGTERM: https://bugs.torproject.org/15435.
 		go func() {
-			if _, err := io.Copy(ioutil.Discard, os.Stdin); err != nil {
-				log.Printf("error copying os.Stdin to ioutil.Discard: %v", err)
+			if _, err := io.Copy(io.Discard, os.Stdin); err != nil {
+				log.Printf("error copying os.Stdin to io.Discard: %v", err)
 			}
 			log.Printf("synthesizing SIGTERM because of stdin close")
 			sigChan <- syscall.SIGTERM
